"""
Configuration schema for Astoria.

Common to all components.
"""
import re
from pathlib import Path
from typing import IO, Optional

from pydantic import BaseModel, validator
from toml import load


class MQTTBrokerInfo(BaseModel):
    """MQTT Broker Information."""

    host: str
    port: int
    enable_tls: bool = False
    topic_prefix: str = "astoria"
    force_protocol_version_3_1: bool = False

    class Config:
        """Pydantic config."""

        extra = "forbid"


KIT_VERSION_REGEX = re.compile(r"^(?P<epoch>\d+)\.(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?P<dev>dev)?(?::(?P<hash>[0-9a-f]{5,40})(?:@(?P<branch>\w+))?)?$")  # noqa: E501


class KitInfo(BaseModel):
    """Kit Information."""

    name: str
    version: str

    @validator('version')
    def version_must_match_regex(cls, v: str) -> str:
        """Validate that the version matches the regex."""
        if not KIT_VERSION_REGEX.match(v):
            raise ValueError("version does not match format")
        return v

    class Config:
        """Pydantic config."""

        extra = "forbid"


class WiFiInfo(BaseModel):
    """System settings for WiFi."""

    interface: str
    bridge: str
    enable_wpa3: bool

    class Config:
        """Pydantic config."""

        extra = "forbid"


class SystemInfo(BaseModel):
    """System settings that don't find elsewhere."""

    cache_dir: Path

    class Config:
        """Pydantic config."""

        extra = "forbid"


class AstoriaConfig(BaseModel):
    """Config schema for Astoria."""

    mqtt: MQTTBrokerInfo
    kit: KitInfo
    wifi: WiFiInfo
    system: SystemInfo

    class Config:
        """Pydantic config."""

        extra = "forbid"

    @classmethod
    def _get_config_path(cls, config_str: Optional[str] = None) -> Path:
        """Check for a config file or search the filesystem for one."""
        CONFIG_SEARCH_PATHS = [
            Path("astoria.toml"),
            Path("/etc/astoria.toml"),
        ]
        if config_str is None:
            for path in CONFIG_SEARCH_PATHS:
                if path.exists() and path.is_file():
                    return path
        else:
            path = Path(config_str)
            if path.exists() and path.is_file():
                return path
        raise FileNotFoundError("Unable to find config file.")

    @classmethod
    def load(cls, config_str: Optional[str] = None) -> 'AstoriaConfig':
        """Load the config."""
        config_path = cls._get_config_path(config_str)
        with config_path.open("r") as fh:
            return cls.load_from_file(fh)

    @classmethod
    def load_from_file(cls, fh: IO[str]) -> 'AstoriaConfig':
        """Load the config from a file."""
        return cls(**load(fh))



"""Types for astmetad."""
import platform
from enum import Enum
from typing import Optional

from pydantic import BaseModel

from astoria import __version__
from astoria.common.config import AstoriaConfig

from .base import ManagerMessage


class RobotMode(Enum):
    """Running Status of the manager daemon."""

    COMP = "COMP"
    DEV = "DEV"


class Metadata(BaseModel):
    """Astoria Metadata."""

    class Config:
        """Pydantic config."""

        validate_assignment = True

    @classmethod
    def init(cls, config: AstoriaConfig) -> 'Metadata':
        """
        Initialise the metadata and populate with data.

        Based on information from software and the astoria configuration.
        """
        uname = platform.uname()
        return cls(
            astoria_version=__version__,
            kernel_version=uname.release,
            arch=uname.machine,
            python_version=platform.python_version(),
            libc_ver="".join(platform.libc_ver()),
            kit_name=config.kit.name,
            kit_version=config.kit.version,
        )

    def is_wifi_valid(self) -> bool:
        """
        Check if the WiFi configuration is valid to be turned on.

        Checks that the WiFi is enabled and that the config params are set.Optional
        :return: boolean indicating whether the WiFi can be turned on.
        """
        return all([
            self.wifi_enabled,
            self.wifi_ssid is not None,
            self.wifi_psk is not None,
            self.wifi_region is not None,
        ])

    # From Meta USB
    arena: str = "A"
    zone: int = 0
    mode: RobotMode = RobotMode.DEV
    marker_offset: int = 0
    game_timeout: Optional[int] = None
    wifi_enabled: bool = True

    # From Software
    astoria_version: str
    kernel_version: str
    arch: str
    python_version: str
    libc_ver: str

    # From astoria.toml
    kit_name: str
    kit_version: str
    wifi_ssid: Optional[str] = None
    wifi_psk: Optional[str] = None
    wifi_region: Optional[str] = None


class MetadataManagerMessage(ManagerMessage):
    """
    Status message for Metadata Manager.

    Published to /astoria/astmetad
    """

    metadata: Metadata



"""Schema, validation and comparison for Kit Bundle files."""
from typing import List

from pydantic import BaseModel

from .config import KIT_VERSION_REGEX, KitInfo


class IncompatibleKitVersionException(Exception):
    """The kit version was incompatible with the bundle version."""


class BundleInfo(BaseModel):
    """Code Bundle Info Schema."""

    version: str


class WiFiInfo(BaseModel):
    """Code Bundle WiFi Information schema."""

    ssid: str
    psk: str
    enabled: bool
    region: str


class CodeBundle(BaseModel):
    """Schema for code bundle information."""

    bundle: BundleInfo
    kit: KitInfo
    wifi: WiFiInfo

    class Config:
        """Pydantic config."""

        extra = "forbid"

    def check_kit_version_is_compatible(self, system_kit_info: KitInfo) -> List[str]:
        """
        Check that the kit version is compatible.

        :returns: Message to show to the user
        :raises IncompatibleKitVersionException:
        """
        if system_kit_info.name != self.kit.name:
            raise IncompatibleKitVersionException(
                f"Code Bundle was for {self.kit.name}, but the kit is "
                f"{system_kit_info.name}. Refusing to start code as software may"
                f" not be compatible.",
            )

        system_match = KIT_VERSION_REGEX.match(system_kit_info.version)
        bundle_match = KIT_VERSION_REGEX.match(self.kit.version)

        bundle_warnings: List[str] = []

        # These are technically unreachable as we are already validating that
        # the version strings match the regex in the schema. Thus these two
        # checks existing mainly for satisfying the type checker and can be
        # safely ignored from test coverage.
        if system_match is None:  # pragma: no cover
            raise RuntimeError("Invalid system kit version")
        if bundle_match is None:  # pragma: no cover
            raise RuntimeError("Invalid bundle kit version")

        if system_match['epoch'] != bundle_match['epoch']:
            raise IncompatibleKitVersionException(
                "Code Bundle was built for a different version of the kit.",
            )

        if system_match['dev']:
            bundle_warnings.append("WARNING: Running on DEVELOPMENT BUILD")

        if system_match['major'] != bundle_match['major'] or \
                system_match['minor'] != bundle_match['minor']:
            bundle_warnings.append(
                "⚠ Your kit software is unsupported and requires an "
                "update. Please update the kit software.",
            )
        elif system_match['patch'] != bundle_match['patch']:
            bundle_warnings.append(
                "⚠ Code Bundle was made for a version of the kit software"
                " which is different than the current version.",
            )
        return bundle_warnings



"""Lifecycle classes to load metadata from disks."""

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from json import JSONDecodeError, loads
from typing import Dict
from zipfile import BadZipFile, ZipFile

import toml
from pydantic import ValidationError

from astoria.common.bundle import CodeBundle
from astoria.common.messages.astdiskd import DiskInfo, DiskUUID

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class AbstractMetadataDiskLifecycle(metaclass=ABCMeta):
    """Load and validate metadata from a disk."""

    def __init__(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        self._uuid = uuid
        self._disk_info = disk_info

        self._diff = self.extract_diff_data()

    @abstractmethod
    def extract_diff_data(self) -> Dict[str, str]:
        """Extract the diff data from the disk."""
        raise NotImplementedError

    @property
    def diff_data(self) -> Dict[str, str]:
        """The data to be used as override."""
        return self._diff


class MetadataDiskLifecycle(AbstractMetadataDiskLifecycle):
    """Load and validate metadata from a JSON file on the disk."""

    def extract_diff_data(self) -> Dict[str, str]:
        """
        Extract the diff data from the disk.

        Loads astoria.json from the disk and parses it as JSON.
        """
        metadata_file_path = self._disk_info.mount_path / "astoria.json"

        try:
            with metadata_file_path.open("r") as fh:
                return loads(fh.read())  # type: ignore
        except FileNotFoundError:
            LOGGER.warning("Unable to find metadata.json.")
        except JSONDecodeError:
            LOGGER.warning("Invalid JSON in astoria.json")
        return {}


class BundleDiskLifecycle(AbstractMetadataDiskLifecycle):
    """Load and validate metadata from a usercode bundle on the disk."""

    def extract_diff_data(self) -> Dict[str, str]:
        """
        Extract the diff data from the disk.

        Loads bundle.toml fron inside the robot.zip
        """
        bundle_path = self._disk_info.mount_path / "robot.zip"

        try:
            with ZipFile(bundle_path) as zf:
                bundle_file = zf.read("bundle.toml")
            bundle_contents = toml.loads(bundle_file.decode())
            bundle = CodeBundle(**bundle_contents)

            return {
                "wifi_ssid": bundle.wifi.ssid,
                "wifi_psk": bundle.wifi.psk,
                "wifi_region": bundle.wifi.region,
                "wifi_enabled": str(bundle.wifi.enabled),
            }
        except FileNotFoundError:
            LOGGER.warning("Unable to find metadata.json.")
        except BadZipFile:
            LOGGER.warning("Bad robot.zip")
        except toml.TomlDecodeError:
            LOGGER.warning("Invalid code bundle.toml")
        except ValidationError:
            LOGGER.warning("Invalid code bundle.toml")
        return {}
